/*
  LCMXO2280.tgt_type       = lscc
  LCMXO2280.device_type    = 0x128D
  LCMXO2280.address_size   = 1116
  LCMXO2280.data_width     = 436
  LCMXO2280.boundary_width = 544

  LFXP2_5E.tgt_type       = lscc
  LFXP2_5E.device_type    = 0x1299
  LFXP2_5E.address_size   = 1938
  LFXP2_5E.data_width     = 636
  LFXP2_5E.boundary_width = 394

  LCMXO2_1200.tgt_type       = lscc
  LCMXO2_1200.device_type    = 0x12B2
  LCMXO2_1200.address_size   = 333
  LCMXO2_1200.data_width     = 1080
  LCMXO2_1200.boundary_width = 208
  LCMXO2_1200.flash_col      = 128
  LCMXO2_1200.flash_cfm_row  = 2175
  LCMXO2_1200.flash_ufm_row  =  512
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>

#include "util.h"
#include "cable.h"
#include "jtag.h"
#include "target.h"
#include "tgtlscc.h"

#define DEV_TYPE_XP2	0x1299
#define DEV_TYPE_XO	0x128D
#define DEV_TYPE_XO2	0x12B2
#define DEV_VENDOR_LSCC	0x043

#define id_to_vendor(id)	((id) & 0xfff)
#define id_to_type(id)		(((id) >> 12) & 0xffff)

struct supported_device {
	int device_type;
	int address_size;
	int data_width;
	int boundary_width; 
	int flash_col; 
	int flash_cfm_row; 
	int flash_ufm_row; 
	char *device_name;
};
#if 0
static struct supported_device supported_device[] = {
 { DEV_TYPE_XP2 , 1938 ,  638 ,  394 ,   0,     0,    0, "LFXP2_5E"},
 { DEV_TYPE_XO  , 1116 ,  436 ,  544 ,   0,     0,    0, "LCMXO2280"}, 
 { DEV_TYPE_XO2 ,  333 , 1080 ,  208 , 128,  2175,  512, "LCMXO2_1200"}, 
 { DEV_TYPE_XO2 ,  420 , 1272 ,  424 , 128,  5120,  512, "LCMXO2_2000"}, 
 { DEV_TYPE_XO2 ,  623 , 1560 ,  552 , 128,  6144,  512, "LCMXO2_4000"}, 
 { DEV_TYPE_XO2 ,  770 , 1992 ,  664 , 128, 16384,  512, "LCMXO2_7000"}, 
 { 0  , 0 , 0 , 0 , "" }
};
#endif
static struct supported_device cur_device;

/*
 DATA FORAMT
 LFXP2_5E :
	      0 - 1236443 bit (1938 * 638) Config data
	1236444 - 1236475 (32)             SED_CRC 
	1236476 - 1236508 (32)             USERCODE

 LCMXO2280:
	     0 - 486575 (1116 * 436)       Config data
	486575 - 486608 (32)               USERCODE
 LCMXO2_1200 (SRAM):
	     0 - 359639  (333 * 1080)      Config data
	359640 - 359671  (32)              USERCODE
 LCMXO2_1200 (FLASH):
	     0 - 278399  (128 * 2175)      Config data(CFM)
	278400 - 343935  (128 * 512)       UFM
	343936 - 344015  (80)              FEATURE
	344016 - 344047  (32)              USERCODE
*/
static int sel_channel (TARGET tgt, CABLE cbl, int chan) {
	int ret = -1;
	int r;

	jtag_init(cbl);
	if (chan == 1) {
		jtag_SIR32(cbl, LSCC_IPA, 8, 1, &r);
	} else if (chan == 2) {
		jtag_SIR32(cbl, LSCC_IPB, 8, 1, &r);
	} else {
		return ret;
	}
	return 0;
}

static int erase_check_xo(TARGET tgt, CABLE cbl) {
	int ret = -1;
	int i,r;
	for (i=0; i<100; i++) {
		jtag_IDLE(cbl, 100, 5); // 100ms idle
		jtag_SDR32(cbl, 0, 1, 1, &r);
		if (r == 1) return 0; // OK
	}
	return ret;
}

static int write_check_xo(TARGET tgt, CABLE cbl) {
	int ret = -1;
	int i,r;
	for (i=0; i<10; i++) {
		jtag_IDLE(cbl, 1, 5); // 1ms idle
		jtag_SDR32(cbl, 0, 1, 1, &r);
		if (r == 1) return 0; // OK
	}
	return ret;
}

static int status_check_xo(TARGET tgt, CABLE cbl) {
	int ret = -1;
	int r;
	jtag_SIR32(cbl, READ_STATUS, 8, 0, NULL);
	jtag_IDLE(cbl, 1, 5); // 1ms idle
	jtag_SDR32(cbl, 0, 1, 1, &r);
	if (r == 0) return 0; // OK
	return ret;
}

static int read_status_xp2(TARGET tgt, CABLE cbl) {
	int r;
	jtag_SIR32(cbl, READ_STATUS, 8, 0, NULL);
	jtag_IDLE(cbl, 1, 5); // 1ms idle
	jtag_SDR32(cbl, 0, 8, 1, &r);
	return r;
}

static int program_status_check_xp2(TARGET tgt, CABLE cbl) {
	int ret = -1;

	int r,i;
	jtag_SIR32(cbl, ISC_PROGRAM_STATUS, 8, 0, NULL);
	for (i=0; i<10; i++) {
	    jtag_IDLE(cbl, 1, 5); // 1ms idle
	    jtag_SDR32(cbl, 0, 1, 1, &r);
	    if (r == 1) return 0; // OK
	}
	return ret;
}

static int xo_enter_conf(TARGET tgt, CABLE cbl) {
	int ret = -1;
	uint32_t device_id;
	int r;
	int is_xo2;
	uint8_t buf[512];
	
	is_xo2 = (id_to_type(tgt->device_id) == DEV_TYPE_XO2);

	if (is_xo2) {
		jtag_SIR32(cbl, XO2_IDCODE, 8, 1, &r);
		jtag_SDR32(cbl, 0xffffffff, 32, 1, &device_id);
	} else {
		jtag_SIR32(cbl, LSCC_IDCODE, 8, 1, &r);
		jtag_SDR32(cbl, 0, 32, 1, &device_id);
	}
	if (tgt->device_id != device_id) {
		if (verbose)
			fprintf(stderr, "   device_id mismatch\n");
		return ret;
	}
	jtag_SIR32(cbl, PRELOAD, 8, 0, NULL);
	memset(buf, 0xff, (tgt->boundary_width+7)/8);
	jtag_SDR(cbl, buf, tgt->boundary_width, 0, NULL);

	return 0;
}

static void xo_exit_conf(TARGET tgt, CABLE cbl, int is_program) {
	int r;
	if (is_program) {
		jtag_SIR32(cbl, ISC_PROGRAM_DONE, 8, 0, NULL);
		jtag_IDLE(cbl, 50, 5); // 50ms idle
		jtag_SIR32(cbl, BYPASS_8, 8, 1, &r);
		if (status_check_xo(tgt, cbl)) {
			if (verbose)
				fprintf(stderr, "   erase check error\n");
			 goto err;
		}
	}
	jtag_SIR32(cbl, ISC_DISABLE, 8, 0, NULL);
	jtag_IDLE(cbl, 1, 5); // 1ms idle
err:
	jtag_SIR32(cbl, BYPASS_8, 8, 1, &r);
	jtag_IDLE(cbl, 1, 5); // 1ms idle
}

static int __erase_flash_xo(TARGET tgt, CABLE cbl, int is_xo) {
	int ret = -1;
	if (!is_xo) {
		if (read_status_xp2(tgt, cbl) & 0x10) goto err;
		if (read_status_xp2(tgt, cbl) & 0x40) goto err;
		if (read_status_xp2(tgt, cbl) & 0x04) goto err;
		jtag_SIR32(cbl, ISC_ERASE_DONE, 8, 0, NULL);
		jtag_IDLE(cbl, 10, 5); // 10ms idle

		jtag_SIR32(cbl, ISC_ERASE, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle
		
		if (program_status_check_xp2(tgt, cbl)) goto err;

		jtag_SIR32(cbl, ISC_DISABLE, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle

		jtag_SIR32(cbl, ISC_ENABLE, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle
		if (read_status_xp2(tgt, cbl) & 0x01) goto err;
		if (read_status_xp2(tgt, cbl) & 0x03) goto err;
	} else {
		jtag_SIR32(cbl, ISC_SRAM_ENABLE, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle

		jtag_SIR32(cbl, ISC_ERASE, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle

		jtag_SIR32(cbl, ISC_ENABLE, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle

		jtag_SIR32(cbl, ISC_ERASE, 8, 0, NULL);
		if (erase_check_xo(tgt, cbl)) {
			if (verbose)
				fprintf(stderr, "   erase check error\n");
			goto err;
		}
		if (status_check_xo(tgt, cbl)) {
			if (verbose)
				fprintf(stderr, "   erase check error\n");
			 goto err;
		}
	}
	return 0;
err:
	return ret;
}

#define VERIFY_MODE	1
#define READ_MODE	0

static int __read_xo(TARGET tgt, CABLE cbl, uint8_t *data, int verify) {
	uint32_t wk;
	uint32_t r,i,w;
	int is_xo;
	int len;
	uint8_t buf[512];

	is_xo = (id_to_type(tgt->device_id) == DEV_TYPE_XO);
	
	r = (tgt->address_size * tgt->data_width +7)/ 8;
	len = r + 4;

	if (verify) {
		progress_start("Verifying", len);
	} else {
		progress_start("Reading", len);
	}

	jtag_SIR32(cbl, LSCC_RESET_ADDRESS, 8, 0, NULL);
	jtag_IDLE(cbl, 1, 5); // 1ms idle
	if (is_xo) {
		jtag_SIR32(cbl, BYPASS_8, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle
	}

	w = tgt->data_width;
	jtag_SIR32(cbl, LSCC_READ_INCR_RTI, 8, 0, NULL);
	for (i=0; i < tgt->address_size ; i++) {
	    jtag_IDLE(cbl, 1, 5); // 1ms idle
	    memset(buf, 0xff, (w+7)/8);
	    if (verify) {
	        jtag_SDR(cbl, buf, w, 1, buf);
		if (bitcmp(buf, 0, data, i * w, w)) { 
			goto err;
		}
	    } else {
	        jtag_SDR_EX(cbl, buf, 0, w, 1, data, i * w);
	    }
	    progress(i * w/8);
	}

	jtag_SIR32(cbl, LSCC_USERCODE, 8, 0, NULL);
	jtag_SDR32(cbl, 0xffffffff, 32, 1, &wk);
	jtag_IDLE(cbl, 1, 5); // 1ms idle
	if (verify) {
		if (tgt->user_id != wk) {
			if (verbose) {
				fprintf(stderr, "   user_id unmached 0x%08x 0x%08x\n"
					,tgt->user_id , wk);
			}
			//goto err;
		}
	} else {
		data[len - 4] = wk;
		data[len - 3] = wk >> 8;
		data[len - 2] = wk >> 16;
		data[len - 1] = wk >> 24;
	}

	if (!is_xo) {
		jtag_SIR32(cbl, READ_SED_CRC, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle
		jtag_SDR32(cbl, 0, 32, 1, &wk);
		if (verify) {
		    if (tgt->crc_data != wk) {
			if (verbose) {
				fprintf(stderr, "   crc_data unmached 0x%08x 0x%08x\n"
					,tgt->crc_data , wk);
			}
			goto err;
		    }
		} else {
			len += 4;
			data[len - 4] = wk;
			data[len - 3] = wk >> 8;
			data[len - 2] = wk >> 16;
			data[len - 1] = wk >> 24;
		}
	}
	progress_stop();
	return 0;
err:
	progress_stop();
	return len;
}

static int erase_flash_xo(TARGET tgt, CABLE cbl) {
	int ret = -1;
	int is_xo;

	if (xo_enter_conf(tgt, cbl)) {
		return ret;
	}
	is_xo = (id_to_type(tgt->device_id) == DEV_TYPE_XO);

	jtag_SIR32(cbl, ISC_ENABLE, 8, 0, NULL);
	jtag_IDLE(cbl, 1, 5); // 1ms idle

	if (__erase_flash_xo(tgt, cbl, is_xo)) {
		goto err;
	}
	xo_exit_conf(tgt, cbl, 0);
	return 0;
err:
	return ret;
}

static int write_xo(TARGET tgt, CABLE cbl, uint8_t *data, int len, int ram) {
	int ret = -1;
	uint32_t r,i,w;
	int is_xo;

	if (xo_enter_conf(tgt, cbl)) {
		return ret;
	}
	is_xo = (id_to_type(tgt->device_id) == DEV_TYPE_XO);
	// size check
	r = (tgt->address_size * tgt->data_width +7)/ 8;
	if (is_xo && (len == r + 4)) {
		tgt->user_id = data[r] 
			| ((uint32_t)data[r+1]<<8) 
			| ((uint32_t)data[r+2]<<16)
			| ((uint32_t)data[r+3]<<24);
	} else if (!is_xo && ( len == r + 8)) {
		tgt->user_id = data[r] 
			| ((uint32_t)data[r+1]<<8) 
			| ((uint32_t)data[r+2]<<16)
			| ((uint32_t)data[r+3]<<24);
		tgt->crc_data = data[r+4] 
			| ((uint32_t)data[r+5]<<8) 
			| ((uint32_t)data[r+6]<<16)
			| ((uint32_t)data[r+7]<<24);
	} else {
		if (verbose)
			fprintf(stderr, "   size mismatch error need = %d len = %d\n",
			r + (is_xo?4:8), len);
		return ret;
	}
	if (verbose) {
		fprintf(stderr, "   user_id  = 0x%08x\n", tgt->user_id);
		if (!is_xo)
		    fprintf(stderr, "   crc_data = 0x%08x\n", tgt->crc_data);
	}

	if (ram) {
		jtag_SIR32(cbl, ISC_SRAM_ENABLE, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle

		jtag_SIR32(cbl, ISC_ERASE, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle
	} else {
		jtag_SIR32(cbl, ISC_ENABLE, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle

		if (__erase_flash_xo(tgt, cbl, is_xo)) {
			if (verbose)
				fprintf(stderr, "   erase error\n");
			return ret;
		}
	}

	progress_start("Programming", len);
	/* program */
	jtag_SIR32(cbl, INIT_ADDRESS, 8, 0, NULL);
	jtag_IDLE(cbl, 1, 5); // 1ms idle
	if (!ram) {
		jtag_SIR32(cbl, BYPASS_8, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle
	}

	w = tgt->data_width;
	if (is_xo) {
	   for (i=0; i < tgt->address_size ; i++) {
		jtag_SIR32(cbl, DATA_SHIFT, 8, 0, NULL);
		jtag_SDR_EX(cbl, data, i * w, w, 0, NULL, 0);
		jtag_SIR32(cbl, LSCC_PROGRAM_INCR_RTI, 8, 0, NULL);
		if (ram) {
	   		jtag_IDLE(cbl, 1, 5); // 1ms idle
		} else if (write_check_xo(tgt, cbl)) {
			goto err_p;
		}
		progress(i * w/8);
	   }
	   if (!ram) {
	        jtag_SIR32(cbl, INIT_ADDRESS, 8, 0, NULL);
	        jtag_IDLE(cbl, 1, 5); // 1ms idle
		jtag_SIR32(cbl, BYPASS_8, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle
	   }
	   jtag_SIR32(cbl, LSCC_USERCODE, 8, 0, NULL);
	   jtag_SDR32(cbl, tgt->user_id, 32, 0, NULL);
	   jtag_SIR32(cbl, ISC_PROGRAM_USERCODE, 8, 0, NULL);
	   jtag_IDLE(cbl, 10, 5); // 10ms idle
	   if (status_check_xo(tgt, cbl)) goto err_p;
	} else {
	   if (!ram) {
		for (i=0; i < tgt->address_size ; i++) {
		    jtag_SIR32(cbl, LSCC_PROGRAM_INCR_RTI, 8, 0, NULL);
		    jtag_SDR_EX(cbl, data, i * w, w, 0, NULL, 0);
		    jtag_IDLE(cbl, 1, 5); // 1ms idle
		    if (program_status_check_xp2(tgt, cbl)) goto err_p;
		    progress(i * w/8);
		}
	    } else {
		jtag_SIR32(cbl, LSCC_PROGRAM_INCR_RTI, 8, 0, NULL);
		for (i=0; i < tgt->address_size ; i++) {
		    jtag_SDR_EX(cbl, data, i * w, w, 0, NULL, 0);
		    jtag_IDLE(cbl, 10, 5); // 10ms idle
		    progress(i * w/8);
		}
	    }
	    jtag_SIR32(cbl, INIT_ADDRESS, 8, 0, NULL);
	    jtag_IDLE(cbl, 1, 5); // 1ms idle
	    jtag_SIR32(cbl, ISC_PROGRAM_USERCODE, 8, 0, NULL);
	    jtag_SDR32(cbl, tgt->user_id, 32, 0, NULL);
	    if (program_status_check_xp2(tgt, cbl)) goto err_p;
	    if (read_status_xp2(tgt, cbl) & 0x01) goto err_p;

	    jtag_SIR32(cbl, PROGRAM_SED_CRC, 8, 0, NULL);
	    jtag_SDR32(cbl, tgt->user_id, 32, 0, NULL);
	    if (program_status_check_xp2(tgt, cbl)) goto err_p;
	}
	progress_stop();

	if (__read_xo(tgt, cbl, data, VERIFY_MODE) ) {
		goto err;
	}
	xo_exit_conf(tgt, cbl, 1);
	return 0;
err_p:
	progress_stop();
err:
	xo_exit_conf(tgt, cbl, 0);
	return ret;
}

static int read_xo(TARGET tgt, CABLE cbl, uint8_t *data, int len, int ram) {
	int ret = -1;
	uint32_t r;
	int is_xo;

	if (xo_enter_conf(tgt, cbl)) {
		return ret;
	}
	is_xo = (id_to_type(tgt->device_id) == DEV_TYPE_XO);
	
	// size check
	// 1116 x 436 = 486576 (bit)
	r = tgt->address_size * tgt->data_width / 8;
	if (verbose)
		fprintf(stderr,"   address_size %d data_width = %d\n",
			tgt->address_size , tgt->data_width);
	if ( len < r + (is_xo?4:8)) {
		if (verbose)
			fprintf(stderr,"   size is short need = %d len = %d\n", r+4,len);
		return ret;
	}
	len = r + 4;

	if (ram && !is_xo) {
		jtag_SIR32(cbl, ISC_SRAM_ENABLE, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle
		jtag_SIR32(cbl, LSCC_READ_CONTROL, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle
		jtag_SDR32(cbl, 0, 32, 0, NULL);
	} else  if (ram && is_xo) {
		jtag_SIR32(cbl, ISC_SRAM_ENABLE, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle
	} else  {
		jtag_SIR32(cbl, ISC_ENABLE, 8, 0, NULL);
		jtag_IDLE(cbl, 1, 5); // 1ms idle
	}
	__read_xo(tgt, cbl, data, READ_MODE);
	xo_exit_conf(tgt, cbl, 0);
	return len;
}

static int read_conf_xo(TARGET tgt, CABLE cbl, uint8_t *data, int len) {
	return read_xo(tgt, cbl, data, len, 1);
}

static int read_flash_xo(TARGET tgt, CABLE cbl, uint8_t *data, int len) {
	return read_xo(tgt, cbl, data, len, 0);
}

static int write_conf_xo(TARGET tgt, CABLE cbl, uint8_t *data, int len) {
	return write_xo(tgt, cbl, data, len, 1);
}

static int write_flash_xo(TARGET tgt, CABLE cbl, uint8_t *data, int len) {
	return write_xo(tgt, cbl, data, len, 0);
}

static int write_conf_dmy(TARGET tgt, CABLE cbl, uint8_t *buf, int len) {
	int ret = -1;
	return ret;
}

static int read_conf_dmy(TARGET tgt, CABLE cbl, uint8_t *buf, int len) {
	int ret = -1;
	return ret;
}

extern char *conf_getenv(char *name);

static int load_conf(char *dev_name, struct supported_device *dev) {
	char buf[256];
	char *p;

fprintf(stderr, "load_conf %s\n", dev_name);
fflush(stderr);
	strcpy(buf, dev_name);
	strcat(buf, ".device_type");
	p = conf_getenv(buf);
	if (!p) {
fprintf(stderr, "  -- device_type undefined\n");
fflush(stderr);
		return -1;
	}
	if (!strncmp("0x", p, 2)) sscanf(p+2,"%x", &dev->device_type);
	else  sscanf(p,"%d", &dev->device_type);
fprintf(stderr, "  -- device_type %04x\n",dev->device_type);
fflush(stderr);

	strcpy(buf, dev_name);
	strcat(buf, ".boundary_width");
	p = conf_getenv(buf);
	if (!p) {
fprintf(stderr, "  -- boudary_width undefined\n");
fflush(stderr);
		return -1;
	}
	if (!strncmp("0x", p, 2)) sscanf(p+2,"%x", &dev->boundary_width);
	else  sscanf(p,"%d", &dev->boundary_width);

	strcpy(buf, dev_name);
	strcat(buf, ".address_size");
	p = conf_getenv(buf);
	if (p) {
	    if (!strncmp("0x", p, 2)) sscanf(p+2,"%x", &dev->address_size);
	    else  sscanf(p,"%d", &dev->address_size);
	}

	strcpy(buf, dev_name);
	strcat(buf, ".data_width");
	p = conf_getenv(buf);
	if (p) {
	    if (!strncmp("0x", p, 2)) sscanf(p+2,"%x", &dev->data_width);
	    else  sscanf(p,"%d", &dev->data_width);
	}

	strcpy(buf, dev_name);
	strcat(buf, ".flash_col");
	p = conf_getenv(buf);
	if (p) {
	    if (!strncmp("0x", p, 2)) sscanf(p+2,"%x", &dev->flash_col);
	    else  sscanf(p,"%d", &dev->flash_col);
	}

	strcpy(buf, dev_name);
	strcat(buf, ".flash_cfm_row");
	p = conf_getenv(buf);
	if (p) {
	    if (!strncmp("0x", p, 2)) sscanf(p+2,"%x", &dev->flash_cfm_row);
	    else  sscanf(p,"%d", &dev->flash_cfm_row);
	}

	strcpy(buf, dev_name);
	strcat(buf, ".flash_ufm_row");
	p = conf_getenv(buf);
	if (p) {
	    if (!strncmp("0x", p, 2)) sscanf(p+2,"%x", &dev->flash_ufm_row);
	    else  sscanf(p,"%d", &dev->flash_ufm_row);
	}
fprintf(stderr, "  -- sucsess\n");
fflush(stderr);
	return 0;
}

int tgtlscc_init (TARGET tgt, CABLE cbl, char *dev_name)  {
	int ret = -1;
	int r;
	struct supported_device *p;
	int dev_type;
//	uint32_t device_id;

	tgt->sel_chan 		= NULL;
	tgt->read_conf 		= NULL;
	tgt->write_conf		= NULL;
	tgt->read_flash 	= NULL;
	tgt->write_flash	= NULL;
	tgt->device_id 		= 0;
	tgt->user_id 		= 0;

	p = &cur_device;
	if (load_conf(dev_name, p)) return -1;
	dev_type = p->device_type;

	tgt->device_name    = p->device_name;
	tgt->boundary_width = p->boundary_width;
	tgt->address_size   = p->address_size;
	tgt->data_width     = p->data_width;
	dev_type = p->device_type;

	jtag_SIR32(cbl, LSCC_IDCODE, 8, 1, &r);
	jtag_SDR32(cbl, 0, 32, 1, &(tgt->device_id));
fprintf(stderr, "  -- device_id %08x \n", tgt->device_id);
fflush(stderr);
	if (id_to_vendor(tgt->device_id) != DEV_VENDOR_LSCC) {
fprintf(stderr, "  -- device_id is not suppoted\n");
fflush(stderr);
		return ret;
	}
	if (id_to_type(tgt->device_id) != dev_type) {
fprintf(stderr, "  -- device_id is not matched\n");
fflush(stderr);
		return ret;
	}
	tgt->sel_chan 	= sel_channel;
	if (id_to_type(tgt->device_id) == DEV_TYPE_XO) {
		tgt->write_conf  	= write_conf_xo;
		tgt->read_conf  	= read_conf_xo;
		// flush ops
		tgt->erase_flash  	= erase_flash_xo;
		tgt->write_flash  	= write_flash_xo;
		tgt->read_flash 	= read_flash_xo;
	} else if (id_to_type(tgt->device_id) == DEV_TYPE_XP2) {
		tgt->read_conf  	= read_conf_xo;
		// flush ops
		tgt->erase_flash  	= erase_flash_xo;
		tgt->read_flash 	= read_flash_xo;
	} else {
		tgt->write_conf  	= write_conf_dmy;
		tgt->read_conf  	= read_conf_dmy;
	}
	return 0;
}
